#import <Cocoa/Cocoa.h>
#import <OpenGL/GL.h>
#import <OpenGL/GLu.h>
#import <QuickTime/QuickTimeComponents.h>

@class XeeView,XeeImageFrame,XeeImageTile,XeeImageLoader;

@interface XeeImage:NSObject
{
	NSString *filename;
	NSDictionary *attrs;

	NSString *format;
	int width,height,depth;
	BOOL transparent;
	NSColor *back;

	BOOL stop;

	XeeView *view;
}

+(XeeImage *)imageForFilename:(NSString *)filename;
+(XeeImage *)imageForFilename:(NSString *)filename view:(XeeView *)targetview;
+(void)registerImageClass:(Class)class;
+(NSArray *)fileTypes;

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes view:(XeeView *)targetview;

-(BOOL)load;
-(BOOL)loaded;
-(void)stopLoading;
-(BOOL)hasBeenStopped;

-(int)frames;
-(void)setFrame:(int)frame;
-(int)frame;

-(BOOL)animated;
-(void)setAnimating:(BOOL)animating;
-(BOOL)animating;

-(void)draw:(NSPoint)point;
-(void)draw:(NSPoint)point size:(NSSize)size;

-(NSString *)filename;
-(int)filesize;
-(NSDate *)date;
-(NSString *)format;
-(int)width;
-(int)height;
-(int)depth;
-(BOOL)transparent;
-(NSColor *)backgroundColor;

-(void)setTransparent:(BOOL)trans;
-(void)setBackgroundColor:(NSColor *)col;

-(NSString *)descriptiveFilesize;
-(NSString *)descriptiveDate;
-(NSString *)descriptiveDepth;

@end



#define XeeImageRGB 0
#define XeeImageYUV 1
#define XeeImageLuma 2

@interface XeeBitmapImage:XeeImage
{
	int type,bytesperrow;
	void *data;

	GLuint texformat,textype;

	GWorldPtr gw;

	NSMutableArray *tiles;
}

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes view:(XeeView *)targetview;
-(id)initWithType:(int)pixeltype width:(int)framewidth height:(int)frameheight view:(XeeView *)targetview;

-(void)dealloc;

-(BOOL)allocBitmapWithType:(int)pixeltype width:(int)framewidth height:(int)frameheight;

-(void)buildTextures;
-(void)buildTexturesRect;
-(void)buildTextures2D;
-(void)updateTextures;
-(void)fixYUVGamma;
-(void)fixYUVGammaFast;
-(void)fixYUVGammaSlow;

-(void)draw:(NSPoint)point size:(NSSize)size;

-(int)type;
-(int)bytesPerRow;
-(void *)data;
-(GWorldPtr)gworld;

@end


@interface XeeImageTile:NSObject
{
	int x,y;
	int width,height;

	GLuint tex,textarget,texformat,textype;
	float texwidth,texheight;

	int fullwidth;
	void *data;
}

-(id)initWithTarget:(GLuint)ttarget format:(GLuint)tformat type:(GLuint)ttype data:(GLubyte *)pixeldata fullwidth:(int)fw fullheight:(int)fh width:(int)w height:(int)h texwidth:(float)tw texheight:(float)th x:(int)x0 y:(int)y0;
-(void)update;
-(void)draw;

@end

