#import "XeeImage.h"
#import "XeeView.h"

#import <GLUT/glut.h>

/*		[view lock];
		while(frame=[enumerator nextObject]) [frame buildTextures];
		[view unlock];*/



@implementation XeeImage

NSMutableArray *imageclasses=nil;

+(XeeImage *)imageForFilename:(NSString *)filename;
{
	return [self imageForFilename:filename view:nil];
}

+(XeeImage *)imageForFilename:(NSString *)filename view:(XeeView *)targetview
{
	NSDictionary *attrs=[[NSFileManager defaultManager] fileAttributesAtPath:filename traverseLink:YES];
	if(!attrs) return nil;

	NSFileHandle *file=[NSFileHandle fileHandleForReadingAtPath:filename];
	if(!file) return nil;

	NSData *block=[file readDataOfLength:512];
	if(!block) return nil;

	[file closeFile];

	NSEnumerator *enumerator=[imageclasses objectEnumerator];
	Class class;

	while(class=[enumerator nextObject])
	{
		XeeImage *image=[[class alloc] initWithFile:filename firstBlock:block attributes:attrs view:targetview];
		if(image) return [image autorelease];
	}

	return nil;
}

+(void)registerImageClass:(Class)class
{
	if(!imageclasses) imageclasses=[[NSMutableArray alloc] initWithCapacity:8];

	[imageclasses addObject:class];
}

+(NSArray *)fileTypes
{
	NSMutableArray *types=[NSMutableArray arrayWithCapacity:32];
	NSEnumerator *enumerator=[imageclasses objectEnumerator];
	Class class;

	while(class=[enumerator nextObject])
	{
		NSEnumerator *typeenum=[[class fileTypes] objectEnumerator];
		NSString *type;
		while(type=[typeenum nextObject]) if(![types containsObject:type]) [types addObject:type];
	}

	return types;
}



-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes view:(XeeView *)targetview
{
	if(self=[super init])
	{
		format=nil;
		width=height=depth=0;
		transparent=NO;
		back=nil;
		stop=NO;

		view=targetview;

		filename=[name retain];
		attrs=[attributes retain];
	}

	return self;
}

-(void)dealloc
{
	[filename release];
	[attrs release];
	[back release];

	[super dealloc];
}

-(BOOL)load { return NO; }

-(BOOL)loaded { return NO; }

-(void)stopLoading { stop=YES; }

-(BOOL)hasBeenStopped { return stop; }

-(int)frames { return 0; }

-(void)setFrame:(int)frame { }

-(int)frame { return 0; }

-(BOOL)animated { return NO; }

-(void)setAnimating:(BOOL)animating { }

-(BOOL)animating { return NO; }

-(void)draw:(NSPoint)point { [self draw:point size:NSMakeSize(width,height)]; }

-(void)draw:(NSPoint)point size:(NSSize)size { }



-(NSString *)filename { return filename; }

-(int)filesize { return [attrs fileSize]; }

-(NSDate *)date { return [attrs fileModificationDate]; }

-(NSString *)format { return format; }

-(int)width { return width; }

-(int)height { return height; }

-(int)depth { return depth; }

-(BOOL)transparent { return transparent; }

-(NSColor *)backgroundColor { return back; }

-(void)setTransparent:(BOOL)trans { transparent=trans; }

-(void)setBackgroundColor:(NSColor *)col { [back release]; back=[col retain]; }

-(NSString *)descriptiveFilesize
{
	int size=[self filesize];

	if(size<10000) return [NSString stringWithFormat:@"%d B",size];
	else if(size<102400) return [NSString stringWithFormat:@"%.2f kB",((float)size/1024.0)];
	else if(size<1024000) return [NSString stringWithFormat:@"%.1f kB",((float)size/1024.0)];
//	else if(size<10240000) return [NSString stringWithFormat:@"%d kB",size/1024];
	else return [NSString stringWithFormat:@"%d kB",size/1024];
}

-(NSString *)descriptiveDate
{
	return [[self date] description];
}

-(NSString *)descriptiveDepth
{
	int d=[self depth];
	if(d<=8) return [NSString stringWithFormat:@"%d colours",1<<d];
	else if(d<=32) return [NSString stringWithFormat:@"%d bits RGB",d]; 
	else if(d<=40) return [NSString stringWithFormat:@"%d greys",1<<(d-32)]; 
	else return @"";
}

@end



// XeeBitmapImage

@implementation XeeBitmapImage

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes view:(XeeView *)targetview
{
	if(self=[super initWithFile:name firstBlock:block attributes:attributes view:targetview])
	{
		type=0;
		bytesperrow=0;
		data=NULL;
		texformat=textype=0;

		gw=NULL;
		tiles=nil;
	}

	return self;
}

-(id)initWithType:(int)pixeltype width:(int)framewidth height:(int)frameheight view:(XeeView *)targetview
{
	if(self=[self initWithFile:nil firstBlock:nil attributes:nil view:targetview])
	{
		if([self allocBitmapWithType:type width:framewidth height:frameheight])
		{
			return self;
		}

		[self release];
	}

	return nil;
}

-(void)dealloc
{
	[view lock];
	[tiles release];
	[view unlock];

	if(gw) DisposeGWorld(gw);

	free(data);

	[super dealloc];
}

-(BOOL)allocBitmapWithType:(int)pixeltype width:(int)framewidth height:(int)frameheight
{
	OSType pixelformat;

	if(data||gw) return NO; // maybe dealloc?

	width=framewidth;
	height=frameheight;
	type=pixeltype;

	switch(type)
	{
		case XeeImageRGB:
			texformat=GL_BGRA;
			textype=GL_UNSIGNED_INT_8_8_8_8_REV;
			bytesperrow=4*width;
			pixelformat=k32ARGBPixelFormat;
		break;

		case XeeImageYUV:
			texformat=GL_YCBCR_422_APPLE;
			textype=GL_UNSIGNED_SHORT_8_8_REV_APPLE;
			bytesperrow=4*((width+1)/2);
			pixelformat=k422YpCbCr8CodecType;
		break;

		case XeeImageLuma: // not implemented yet!
			texformat=GL_LUMINANCE;
			textype=GL_UNSIGNED_BYTE;
			bytesperrow=width;
			pixelformat=0;
		break;
	}

	data=malloc(bytesperrow*height);

	if(data)
	{
		Rect rect;
		SetRect(&rect,0,0,width,height);
		QTNewGWorldFromPtr(&gw,pixelformat,&rect,NULL,NULL,0,data,bytesperrow);

		if(gw)
		{
			return YES;
		}
		free(data);
	}

	return NO;
}

-(void)buildTextures
{
	[view lock];

	if(glutExtensionSupported("GL_EXT_texture_rectangle")) [self buildTexturesRect];
	else [self buildTextures2D];

	[view unlock];
}

-(void)buildTexturesRect
{
	GLint maxtilesize;
	glGetIntegerv(GL_MAX_RECTANGLE_TEXTURE_SIZE_EXT,&maxtilesize);
	if(maxtilesize>512) maxtilesize=512;

	int cols=(width+maxtilesize-1)/maxtilesize;
	int rows=(height+maxtilesize-1)/maxtilesize;

	tiles=[[NSMutableArray alloc] initWithCapacity:rows*cols];

	if(tiles)
	{
		for(int row=0;row<rows;row++)
		{
			int tile_h=(row==rows-1)?(height-(rows-1)*maxtilesize):maxtilesize;

			for(int col=0;col<cols;col++)
			{
				int tile_w=(col==cols-1)?(width-(cols-1)*maxtilesize):maxtilesize;

				[tiles addObject:[[[XeeImageTile alloc] initWithTarget:GL_TEXTURE_RECTANGLE_EXT format:texformat type:textype data:data
				fullwidth:width fullheight:height width:tile_w height:tile_h texwidth:tile_w texheight:tile_h
				x:col*maxtilesize y:row*maxtilesize] autorelease]];
			}
		}
	}
}

-(void)buildTextures2D
{
	GLint maxtilesize;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE,&maxtilesize);
	if(maxtilesize>512) maxtilesize=512;

	int colsize,rowsize;
	int x,y;

	tiles=[[NSMutableArray alloc] initWithCapacity:32];

	if(tiles)
	{
		rowsize=maxtilesize;
		y=0;
		while(y<height)
		{
			while(y+rowsize>height) rowsize/=2;

			colsize=maxtilesize;
			x=0;
			while(x<width)
			{
				while(x+colsize>width) colsize/=2;

				[tiles addObject:[[[XeeImageTile alloc] initWithTarget:GL_TEXTURE_2D format:texformat type:textype data:data
				fullwidth:width fullheight:height width:colsize height:rowsize texwidth:1 texheight:1 x:x y:y] autorelease]];

				x+=colsize;
			}

			y+=rowsize;
		}
	}
}

-(void)updateTextures
{
	[view lock];

	NSEnumerator *enumerator=[tiles objectEnumerator];
	XeeImageTile *tile;

	while(tile=[enumerator nextObject])
	{
		[tile update];
	}

	[view unlock];
}

-(void)draw:(NSPoint)point size:(NSSize)size
{
	glPushMatrix();

	glTranslatef(point.x,point.y,0);
	glScalef(size.width/(float)width,size.height/(float)height,1);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

	NSEnumerator *enumerator=[tiles objectEnumerator];
	XeeImageTile *tile;

	while(tile=[enumerator nextObject])
	{
		[tile draw]; // needs a check
	}

	glPopMatrix();
}

-(void)fixYUVGamma
{
	if([[NSUserDefaults standardUserDefaults] integerForKey:@"jpegGamma"]==0) [self fixYUVGammaFast];
	else [self fixYUVGammaSlow];
}

-(void)fixYUVGammaFast
{
	if(type!=XeeImageYUV) return;

	static const unsigned char gammatable[256]=
	{
		0x00,0x00,0x01,0x01,0x02,0x02,0x03,0x03,0x04,0x04,0x05,0x05,0x06,0x07,0x07,0x08,
		0x09,0x09,0x0a,0x0b,0x0b,0x0c,0x0d,0x0d,0x0e,0x0f,0x10,0x10,0x11,0x12,0x13,0x13,
		0x14,0x15,0x16,0x17,0x17,0x18,0x19,0x1a,0x1b,0x1b,0x1c,0x1d,0x1e,0x1f,0x1f,0x20,
		0x21,0x22,0x23,0x24,0x25,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2c,0x2d,0x2e,
		0x2f,0x30,0x31,0x32,0x33,0x34,0x35,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,
		0x3e,0x3f,0x40,0x41,0x42,0x43,0x44,0x45,0x45,0x46,0x47,0x48,0x49,0x4a,0x4b,0x4c,
		0x4d,0x4e,0x4f,0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5a,0x5b,0x5c,
		0x5d,0x5e,0x5f,0x60,0x61,0x62,0x63,0x64,0x65,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,
		0x6e,0x6f,0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7b,0x7c,0x7d,0x7e,
		0x7f,0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x89,0x8a,0x8b,0x8c,0x8d,0x8e,0x8f,
		0x90,0x91,0x92,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9e,0x9f,0xa0,0xa1,
		0xa2,0xa3,0xa4,0xa5,0xa7,0xa8,0xa9,0xaa,0xab,0xac,0xad,0xaf,0xb0,0xb1,0xb2,0xb3,
		0xb4,0xb5,0xb7,0xb8,0xb9,0xba,0xbb,0xbc,0xbd,0xbf,0xc0,0xc1,0xc2,0xc3,0xc4,0xc6,
		0xc7,0xc8,0xc9,0xca,0xcb,0xcd,0xce,0xcf,0xd0,0xd1,0xd3,0xd4,0xd5,0xd6,0xd7,0xd8,
		0xda,0xdb,0xdc,0xdd,0xde,0xe0,0xe1,0xe2,0xe3,0xe4,0xe6,0xe7,0xe8,0xe9,0xea,0xec,
		0xed,0xee,0xef,0xf0,0xf2,0xf3,0xf4,0xf5,0xf6,0xf8,0xf9,0xfa,0xfb,0xfd,0xfe,0xff,
	};

	int linewidth=4*((width+1)/2);

	for(int y=0;y<height;y++)
	{
		int n=width;
		unsigned char *ptr=(unsigned char *)data+y*linewidth+1;

		while(n--)
		{
			*ptr=gammatable[*ptr];
			ptr+=2;
		}
	}
}

-(void)fixYUVGammaSlow
{
	if(type!=XeeImageYUV) return;

	static const unsigned char croptable[256+2048]=
	{
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x01,0x01,0x02,0x02,0x03,0x03,0x04,0x04,0x05,0x05,0x06,0x07,0x07,0x08,
		0x09,0x09,0x0a,0x0b,0x0b,0x0c,0x0d,0x0d,0x0e,0x0f,0x10,0x10,0x11,0x12,0x13,0x13,
		0x14,0x15,0x16,0x17,0x17,0x18,0x19,0x1a,0x1b,0x1b,0x1c,0x1d,0x1e,0x1f,0x1f,0x20,
		0x21,0x22,0x23,0x24,0x25,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2c,0x2d,0x2e,
		0x2f,0x30,0x31,0x32,0x33,0x34,0x35,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,
		0x3e,0x3f,0x40,0x41,0x42,0x43,0x44,0x45,0x45,0x46,0x47,0x48,0x49,0x4a,0x4b,0x4c,
		0x4d,0x4e,0x4f,0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5a,0x5b,0x5c,
		0x5d,0x5e,0x5f,0x60,0x61,0x62,0x63,0x64,0x65,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,
		0x6e,0x6f,0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7b,0x7c,0x7d,0x7e,
		0x7f,0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x89,0x8a,0x8b,0x8c,0x8d,0x8e,0x8f,
		0x90,0x91,0x92,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9e,0x9f,0xa0,0xa1,
		0xa2,0xa3,0xa4,0xa5,0xa7,0xa8,0xa9,0xaa,0xab,0xac,0xad,0xaf,0xb0,0xb1,0xb2,0xb3,
		0xb4,0xb5,0xb7,0xb8,0xb9,0xba,0xbb,0xbc,0xbd,0xbf,0xc0,0xc1,0xc2,0xc3,0xc4,0xc6,
		0xc7,0xc8,0xc9,0xca,0xcb,0xcd,0xce,0xcf,0xd0,0xd1,0xd3,0xd4,0xd5,0xd6,0xd7,0xd8,
		0xda,0xdb,0xdc,0xdd,0xde,0xe0,0xe1,0xe2,0xe3,0xe4,0xe6,0xe7,0xe8,0xe9,0xea,0xec,
		0xed,0xee,0xef,0xf0,0xf2,0xf3,0xf4,0xf5,0xf6,0xf8,0xf9,0xfa,0xfb,0xfd,0xfe,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
	};

#define SCALEBITS 10
#define ONE_HALF (1<<(SCALEBITS-1))
#define FIX(x) ((int)((x)*(1<<SCALEBITS)+0.5))

	int linewidth=((width+1)/2);
    unsigned const char *cm=croptable+1024;

	for(int y=0;y<height;y++)
	{
		int n=linewidth;
		unsigned int *ptr=(unsigned int *)data+y*linewidth;

		while(n--)
		{
			unsigned int val=*ptr;

			// cb=u, cr=v
			int cb=(val>>24)&0xff;
			int y1=(val>>16)&0xff;
			int cr=(val>>8)&0xff;
			int y2=val&0xff;

			int cb_=cb-128;
			int cr_=cr-128;
			int r_add= FIX(1.40200)*cr_+ONE_HALF;
			int g_add=-FIX(0.34414)*cb_-FIX(0.71414)*cr_+ONE_HALF;
			int b_add= FIX(1.77200)*cb_+ONE_HALF;

			int y1_=y1<<SCALEBITS;
			int r1=cm[(y1_+r_add)>>SCALEBITS];
			int g1=cm[(y1_+g_add)>>SCALEBITS];
			int b1=cm[(y1_+b_add)>>SCALEBITS];

			int y2_=y2<<SCALEBITS;
			int r2=cm[(y2_+r_add)>>SCALEBITS];
			int g2=cm[(y2_+g_add)>>SCALEBITS];
			int b2=cm[(y2_+b_add)>>SCALEBITS];

			int new_y1=(FIX(0.29900)*r1+FIX(0.58700)*g1+FIX(0.11400)*b1+ONE_HALF)>>SCALEBITS;
			int new_y2=(FIX(0.29900)*r2+FIX(0.58700)*g2+FIX(0.11400)*b2+ONE_HALF)>>SCALEBITS;
			r1+=r2; g1+=g2; b1+=b2;
			int new_cb=((-FIX(0.16874)*r1-FIX(0.33126)*g1+FIX(0.50000)*b1+(ONE_HALF<<1)-1)>>(SCALEBITS+1))+128;
			int new_cr=(( FIX(0.50000)*r1-FIX(0.41869)*g1-FIX(0.08131)*b1+(ONE_HALF<<1)-1)>>(SCALEBITS+1))+128;

			*ptr++=(new_cb<<24)|(new_y1<<16)|(new_cr<<8)|new_y2;
		}
	}
}

-(int)frames { return 1; }

-(int)type { return type; }

-(int)bytesPerRow { return bytesperrow; }

-(void *)data { return data; }

-(GWorldPtr)gworld { return gw; }

@end



// XeeImageTile

@implementation XeeImageTile

-(id)initWithTarget:(GLuint)ttarget format:(GLuint)tformat type:(GLuint)ttype data:(GLubyte *)pixeldata fullwidth:(int)fw fullheight:(int)fh width:(int)w height:(int)h texwidth:(float)tw texheight:(float)th x:(int)x0 y:(int)y0
{
	if(self=[super init])
	{
		width=w;
		height=h;
		x=x0;
		y=y0;

		texwidth=tw;
		texheight=th;
		textarget=ttarget;
		texformat=tformat;
		textype=ttype;

		fullwidth=fw;
		data=pixeldata;

		glGenTextures(1,&tex);
		if(tex)
		{
			[self update];

			glTexParameteri(textarget,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexParameteri(textarget,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
			glTexParameteri(textarget,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
			glTexParameteri(textarget,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
			glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);

			if(glGetError()==GL_NO_ERROR)
			{
				return self;
			}
		}

		[self release];
	}

	return nil;
}


-(void)dealloc
{
	if(tex) glDeleteTextures(1,&tex);

	[super dealloc];
}

-(void)update
{
	glBindTexture(textarget,tex);

	glPixelStorei(GL_UNPACK_ROW_LENGTH,fullwidth);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS,x);
	glPixelStorei(GL_UNPACK_SKIP_ROWS,y);
	glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE,GL_TRUE);

//	glTexImage2D(textype,0,GL_RGB8,width,height,0,GL_BGRA,GL_UNSIGNED_INT_8_8_8_8_REV,data); // REV: ARGB -> BGRA
//	glTexImage2D(textype,0,GL_RGBA8,width,height,0,GL_YCBCR_422_APPLE,GL_UNSIGNED_SHORT_8_8_REV_APPLE,buffer);
	glTexImage2D(textarget,0,GL_RGBA8,width,height,0,texformat,textype,data);
}

-(void)draw
{
	if(!tex) return;

	glPushMatrix();
	glTranslatef(x,y,0);

	glEnable(textarget);
	glBindTexture(textarget,tex);

	glBegin(GL_QUADS);

	glTexCoord2f(0,0);
	glVertex2i(0,0);

	glTexCoord2f(0,texheight);
	glVertex2i(0,height);

	glTexCoord2f(texwidth,texheight);
	glVertex2i(width,height);

	glTexCoord2f(texwidth,0);
	glVertex2i(width,0);

	glEnd();
		
	glPopMatrix();
}

@end
