#import "XeeMisc.h"
#import "XeeImage.h"
#import "XeeController.h"
#import "XeeStrings.h"

#define XeeDragDataType @"XeeDragDataType"
#define XeeDestinationUpdate @"XeeDestinationUpdate"



@implementation XeeDestinationView

-(void)awakeFromNib
{
	[self setDataSource:[XeeDestinationListHandler defaultObject]];
	[self setDelegate:[XeeDestinationListHandler defaultObject]];
    [self registerForDraggedTypes:[NSArray arrayWithObjects:XeeDragDataType,NSFilenamesPboardType,nil]];

	[self setDoubleAction:@selector(destinationListClick:)];

	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(updateData:) name:XeeDestinationUpdate object:nil];
}

-(void)mouseDown:(NSEvent *)event
{
	NSPoint click=[self convertPoint:[event locationInWindow] fromView:nil];
	[self selectRowIndexes:[NSIndexSet indexSetWithIndex:[self rowAtPoint:click]] byExtendingSelection:NO];

	[super mouseDown:event];
}

-(void)keyDown:(NSEvent *)event
{
	unichar c=[[event characters] characterAtIndex:0];

	if(c==13) [[NSApplication sharedApplication] sendAction:@selector(destinationListClick:) to:nil from:self];
	else [super keyDown:event];
}

-(unsigned int)draggingSourceOperationMaskForLocal:(BOOL)local
{
    return NSDragOperationDelete;
}

-(void)dragImage:(NSImage *)image at:(NSPoint)loc offset:(NSSize)offset event:(NSEvent *)event pasteboard:(NSPasteboard *)pboard source:(id)src slideBack:(BOOL)slide
{
	[super dragImage:image at:loc offset:offset event:event pasteboard:pboard source:src slideBack:NO];
	outside=NO;
	suggest_outside=NO;
}

-(NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender
{
	outside=NO;
	return [super draggingEntered:sender];
}

-(void)draggingExited:(id <NSDraggingInfo>)sender
{
	suggest_outside=YES;
	[super draggingExited:sender];
}

-(void)draggedImage:(NSImage *)draggedImage movedTo:(NSPoint)screenPoint
{
	if(suggest_outside)
	{
		outside=YES;
		suggest_outside=NO;
	}
}

-(void)draggedImage:(NSImage *)image endedAt:(NSPoint)point operation:(NSDragOperation)op
{
	if(outside&&op!=NSDragOperationEvery)
	{
		NSShowAnimationEffect(NSAnimationEffectDisappearingItemDefault,
		[[self window] convertBaseToScreen:[[self window] mouseLocationOutsideOfEventStream]],NSZeroSize,nil,nil,nil);
		[[[[NSSound alloc] initWithContentsOfFile:@"/System/Library/Components/CoreAudio.component/Contents/Resources/SystemSounds/dock/poof item off dock.aif" byReference:NO] autorelease] play];
		[[self dataSource] deleteLastRows];
	}
}

-(void)updateData:(id)notification
{
	[self reloadData];
}

@end




@implementation XeeDestinationListHandler

-(id)init
{
	if(self=[super init])
	{
		NSArray *destarray=[[NSUserDefaults standardUserDefaults] objectForKey:@"destinations"];
		destinations=[[NSMutableArray alloc] initWithCapacity:16];
		[self addDestinations:destarray index:0];

		lastrowset=nil;
	}
	return self;
}

-(id)tableView:(NSTableView *)table objectValueForTableColumn:(NSTableColumn *)column row:(int)row
{
	NSString *ident=[column identifier];

	if(row==0)
	{
		if([ident isEqual:@"filename"]) return S_CHOOSEDIR;
		else return nil;
	}
	else
	{
		NSMutableDictionary *dict=[destinations objectAtIndex:row-1];

		if([ident isEqual:@"filename"])
		{
			return [[dict objectForKey:@"filename"] lastPathComponent];
		}
		else
		{
			NSImage *image=[dict objectForKey:@"icon"];

			if(!image)
			{
				image=[[NSWorkspace sharedWorkspace] iconForFile:[dict objectForKey:@"filename"]];
				[dict setObject:image forKey:@"icon"];
			}

			return image;
		}
	}
}

-(int)numberOfRowsInTableView:(NSTableView *)table
{
	return [destinations count]+1;
}

-(BOOL)tableView:(NSTableView *)table writeRowsWithIndexes:(NSIndexSet *)rows toPasteboard:(NSPasteboard*)pboard
{
	if([rows containsIndex:0]) return 0;

	[pboard declareTypes:[NSArray arrayWithObject:XeeDragDataType] owner:self];
	[pboard setString:[[destinations objectAtIndex:[rows firstIndex]-1] objectForKey:@"filename"] forType:XeeDragDataType];

	[lastrowset release];
	lastrowset=[rows retain];

	return YES;
}

-(NSDragOperation)tableView:(NSTableView*)table validateDrop:(id <NSDraggingInfo>)info proposedRow:(int)row proposedDropOperation:(NSTableViewDropOperation)op
{
	NSPasteboard *pboard=[info draggingPasteboard];

	if(op!=NSTableViewDropAbove) return NSDragOperationNone;
	if(row==0) return NSDragOperationNone;

	if([[pboard types] containsObject:NSFilenamesPboardType])
	{
		NSArray *files=[pboard propertyListForType:NSFilenamesPboardType];

		NSEnumerator *enumerator=[files objectEnumerator];
		NSString *file;
		while(file=[enumerator nextObject])
		{
			FSRef ref;
			if(CFURLGetFSRef((CFURLRef)[NSURL fileURLWithPath:file],&ref))
			{
				Boolean folder,aliased;
				if(!FSResolveAliasFileWithMountFlags(&ref,TRUE,&folder,&aliased,kResolveAliasFileNoUI))
				{
					if(!folder) return NSDragOperationNone;
				}
			}
		}
	}

	return NSDragOperationEvery;
}

-(BOOL)tableView:(NSTableView*)table acceptDrop:(id <NSDraggingInfo>)info row:(int)row dropOperation:(NSTableViewDropOperation)op
{
	if(op!=NSTableViewDropAbove) return NSDragOperationNone;
	if(row==0) return NSDragOperationNone;

	NSPasteboard *pboard=[info draggingPasteboard];
	NSArray *files;
	int insindex=row-1;

	if([[pboard types] containsObject:NSFilenamesPboardType]) files=[pboard propertyListForType:NSFilenamesPboardType];
	else files=[NSArray arrayWithObject:[pboard stringForType:XeeDragDataType]];

	NSEnumerator *enumerator=[files objectEnumerator];
	NSString *file;
	while(file=[enumerator nextObject])
	{
		int remindex=[self findDestination:file];
		if(remindex!=NSNotFound)
		{
			[self removeDestination:remindex];
			if(remindex<insindex) insindex--;
		}
	}

	[self addDestinations:files index:insindex];

	[self updateTables];
	[table selectRowIndexes:[NSIndexSet indexSetWithIndex:insindex+1] byExtendingSelection:NO];
	[self saveList];

	return YES;
}

//-(float)tableView:(NSTableView *)tableView heightOfRow:(int)row

-(void)deleteLastRows
{
	[self removeDestination:[lastrowset firstIndex]-1];

	[lastrowset release];
	lastrowset=nil;

	[self updateTables];
	[self saveList];
}

-(void)suggestInsertion:(NSString *)directory
{
	if([self findDestination:directory]==NSNotFound) [self addDestinations:[NSArray arrayWithObject:directory] index:[destinations count]];

	[self updateTables];
	[self saveList];
}

-(void)addDestinations:(NSArray *)directories index:(int)index
{
	NSEnumerator *enumerator=[directories reverseObjectEnumerator];
	NSString *directory;
	while(directory=[enumerator nextObject])
	{
		[destinations insertObject:[NSMutableDictionary dictionaryWithObjectsAndKeys:directory,@"filename",nil] atIndex:index];
	}
}

-(int)findDestination:(NSString *)directory
{
	int count=[destinations count];
	for(int i=0;i<count;i++) if([[[destinations objectAtIndex:i] objectForKey:@"filename"] isEqual:directory]) return i;

	return NSNotFound;
}

-(void)removeDestination:(int)index
{
	[destinations removeObjectAtIndex:index];
}

-(NSDictionary *)destination:(int)index
{
	return [destinations objectAtIndex:index];
}

-(void)updateTables
{
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:XeeDestinationUpdate object:nil]];
}

-(void)saveList
{
	NSMutableArray *array=[NSMutableArray arrayWithCapacity:[destinations count]];

	NSEnumerator *enumerator=[destinations objectEnumerator];
	NSDictionary *destination;
	while(destination=[enumerator nextObject]) [array addObject:[destination objectForKey:@"filename"]];

	[[NSUserDefaults standardUserDefaults] setObject:array forKey:@"destinations"];
}

+(XeeDestinationListHandler *)defaultObject
{
	static XeeDestinationListHandler *def=nil;

	if(!def) def=[[XeeDestinationListHandler alloc] init];
	return def;
}

@end



@implementation XeeCloseButton

-(id)initWithFrame:(NSRect)frame
{
	if(self=[super initWithFrame:frame])
	{
		[self setTitle:@""];
		[self setBordered:NO];
		[self setImage:[NSImage imageNamed:@"close_4"]];
		[self setAlternateImage:[NSImage imageNamed:@"close_2"]];
	}
	return self;
}

@end



@implementation XeeCollisionPanel

-(void)run:(NSWindow *)window source:(XeeImage *)src destination:(XeeImage *)dest mode:(int)mode;
{
	srcimage=[src retain];
	destimage=[dest retain];
	transfermode=mode;

	NSImage *iconimage=[[NSWorkspace sharedWorkspace] iconForFile:[destimage filename]];
	[iconimage setSize:NSMakeSize(128,128)];
	[icon setImage:iconimage];
	[titlefield setStringValue:[NSString stringWithFormat:S_FILEEXISTS,[[destimage filename] lastPathComponent]]];

	[oldsize setStringValue:[NSString stringWithFormat:@"%d",[destimage filesize]]];
	[newsize setStringValue:[NSString stringWithFormat:@"%d",[srcimage filesize]]];

	[oldformat setStringValue:[NSString stringWithFormat:@"%dx%d\n%@ %@",
	[destimage width],[destimage height],[destimage descriptiveDepth],[destimage format]]];
	[newformat setStringValue:[NSString stringWithFormat:@"%dx%d\n%@ %@",
	[srcimage width],[srcimage height],[srcimage descriptiveDepth],[srcimage format]]];

	[olddate setStringValue:[destimage descriptiveDate]];
	[newdate setStringValue:[srcimage descriptiveDate]];

	NSString *newname;
	NSString *destdir=[[destimage filename] stringByDeletingLastPathComponent];
	NSString *destname=[[[destimage filename] lastPathComponent] stringByDeletingPathExtension];
	NSString *destext=[[destimage filename] pathExtension];
	int n=1;

	do { newname=[destdir stringByAppendingPathComponent:[[NSString stringWithFormat:@"%@-%d",destname,n++] stringByAppendingPathExtension:destext]]; }
	while([[NSFileManager defaultManager] fileExistsAtPath:newname]);

	[namefield setStringValue:[newname lastPathComponent]];

	[[NSApplication sharedApplication] beginSheet:self modalForWindow:window modalDelegate:nil didEndSelector:nil contextInfo:nil];
}

-(void)cancelClick:(id)sender
{
	[[NSApplication sharedApplication] endSheet:self];
	[self orderOut:nil];

	[srcimage release];
	[destimage release];
}

-(void)renameClick:(id)sender
{
	[[NSApplication sharedApplication] endSheet:self];
	[self orderOut:nil];

	NSString *destination=[[[destimage filename] stringByDeletingLastPathComponent] stringByAppendingPathComponent:[namefield stringValue]];
	[controller attemptToTransferFile:[srcimage filename] to:destination mode:transfermode];

	[srcimage release];
	[destimage release];
}

-(void)replaceClick:(id)sender
{
	[[NSApplication sharedApplication] endSheet:self];
	[self orderOut:nil];

	[controller transferFile:[srcimage filename] to:[destimage filename] mode:transfermode];

	[srcimage release];
	[destimage release];
}


@end
